#define GC_STATS		1
#define GC_MIN_NEW_SPACE	1024

#define GC_INFO			0
#define GC_DEBUG		0

#define GC_SPY			0

#include "gsgc.c"

#define _STRING(X)		#X
#define _SOURCEPOS(F, L, V)	F":"_STRING(L)":"V
#define SOURCEPOS(X)		_SOURCEPOS(__FILE__, __LINE__, #X)

#define GC_PROTECT(GC, VAR)	GC_push_root(GC, (GC_PTR *)&VAR, SOURCEPOS(VAR))

#define VERBOSE 0

long objs= 0, bytes= 0;

#define RAND(N)	({ long n= (1 + (int)((float)N * (rand() / (RAND_MAX + 1.0))));	 bytes += n;  n; })

struct cell
{
    int		 tag;
    struct cell *next;
};

void *mklist(GarbageCollector *gc, int n)
{
    struct cell *cell= 0;
    if (n)
    {
	GC_BEGIN(gc);
	cell= GC_malloc_mapped(gc, sizeof(struct cell), 2);		++objs;	 bytes += sizeof(struct cell);
	GC_PROTECT(gc, cell);
	cell->tag= n;
	cell->next= mklist(gc, n - 1);
	GC_check_store(gc, cell, cell->next);
	GC_END(gc);
    }
    return cell;
}

void delist(struct cell *cell)
{
    if (cell && cell->next && cell->next->next)
    {
	cell->next= cell->next->next;
	delist(cell->next->next);
    }
}

void status(char *prefix, GarbageCollector *gc)
{
    printf("%s: %ld/%ld bytes = %ld/%ld new + %ld/%ld old\n",
	   prefix,
	   Space_used(&gc->newSpace) + Space_used(&gc->oldSpace),
	   Space_size(&gc->newSpace) + Space_size(&gc->oldSpace),
	   Space_used(&gc->newSpace),  Space_size(&gc->newSpace),
	   Space_used(&gc->oldSpace),  Space_size(&gc->oldSpace));
}

char *km(long n)
{
    static char buf[32];
    if      (n < 5 * 1024       )	snprintf(buf, sizeof(buf), "%ld bytes", n);
    else if (n < 5 * 1024 * 1024)	snprintf(buf, sizeof(buf), "%ld Kbytes", n / 1024);
    else				snprintf(buf, sizeof(buf), "%ld Mbytes", n / 1024 / 1024);
    return buf;
}

long kmgtoul(char *s)
{
    char *last;
    long n= strtoul(s, &last, 0);
    switch (*last)
    {
	case 'g': case 'G':	n *= 1024;
	case 'm': case 'M':	n *= 1024;
	case 'k': case 'K':	n *= 1024;
    }
    return n;
}

void run(long heapSize)
{
    GarbageCollector *gc= GC_new(heapSize);
    int i, j;
    void *a, *b, *c, *d, *e;

    objs= 0;
    bytes= 0;

# if (GC_SPY)
    GC_spy(gc, 1);
# endif

    printf("\ninitial heap size is %s\n\n", km(Space_size(&gc->newSpace) + Space_size(&gc->oldSpace)));

    printf("allocating objects...\n");

    for (i= 0;	i < 10000;  ++i)
    {
	GC_BEGIN(gc);

	a= 0;  GC_PROTECT(gc, a);
	b= 0;  GC_PROTECT(gc, b);
	c= 0;  GC_PROTECT(gc, c);
	d= 0;  GC_PROTECT(gc, d);
	e= 0;  GC_PROTECT(gc, e);

#	if (!VERBOSE)
#	  define printf(...)
#	endif

	a= GC_malloc(gc, RAND(1));	printf("%p\n", a);	++objs;
	b= GC_malloc(gc, RAND(10));	printf("%p\n", b);	++objs;
	c= GC_malloc(gc, RAND(100));	printf("%p\n", c);	++objs;
	d= GC_malloc(gc, RAND(1000));	printf("%p\n", d);	++objs;
	e= GC_malloc(gc, RAND(10000));	printf("%p\n", e);	++objs;

	a= GC_malloc(gc, RAND(100));	printf("%p\n", a);	++objs;
	b= GC_malloc(gc, RAND(200));	printf("%p\n", b);	++objs;
	c= GC_malloc(gc, RAND(300));	printf("%p\n", c);	++objs;
	d= GC_malloc(gc, RAND(400));	printf("%p\n", d);	++objs;
	e= GC_malloc(gc, RAND(500));	printf("%p\n", e);	++objs;

	a= GC_malloc(gc, RAND(4));	printf("%p\n", a);	++objs;
	b= GC_malloc(gc, RAND(16));	printf("%p\n", b);	++objs;
	c= GC_malloc(gc, RAND(64));	printf("%p\n", c);	++objs;
	d= GC_malloc(gc, RAND(256));	printf("%p\n", d);	++objs;
	e= GC_malloc(gc, RAND(1024));	printf("%p\n", e);	++objs;

	a= GC_malloc(gc, RAND(713));	printf("%p\n", a);	++objs;
	b= GC_malloc(gc, RAND(713));	printf("%p\n", b);	++objs;
	c= GC_malloc(gc, RAND(713));	printf("%p\n", c);	++objs;
	d= GC_malloc(gc, RAND(713));	printf("%p\n", d);	++objs;
	e= GC_malloc(gc, RAND(713));	printf("%p\n", e);	++objs;

#	undef printf

#	if (SYNC_GC)
	{
	    if (i % 1000 == 0) status("alloc", gc);
	    GC_collect(gc);
	    if (i % 1000 == 0) status("   gc", gc);
	}
#	endif

	GC_END(gc);
    }

    printf("allocating lists...\n");

    {
	GC_BEGIN(gc);
	a= 0;  GC_PROTECT(gc, a);

	for (i= 0;  i < 10;  ++i)
	{
	    for (j= 0;	j < 100;  ++j)
	    {
		a= mklist(gc, 2000);
		delist(a);
#		if (VERBOSE)
		{
		    struct cell *c= a;
		    printf("----\n");
		    while (c)
		    {
			printf("%p %d %p\n", c, c->tag >> 1, c->next);
			c= c->next;
		    }
		}
#		endif
	    }
#	    if (SYNC_GC)
	    {
		status("alloc", gc);
		GC_collect(gc);
		status("   gc", gc);
	    }
#	   endif
	}
	GC_END(gc);
    }

    printf("finished allocating %s in %ld objects\n\n", km(bytes), objs);

    status("alive", gc);
    GC_collect(gc);
    status("   gc", gc);
    GC_collect_all(gc);
    status(" full", gc);

    printf("\n%i collections, %i full collections, ", gc->seq, gc->full);
    printf("final heap size is %s\n", km(Space_size(&gc->newSpace) + Space_size(&gc->oldSpace)));

    GC_delete(gc);

    printf("\n");
}

int main(int argc, char **argv)
{
    if (1 == argc)
	run(4*1024*1024);
    else
	while (--argc)
	    run(kmgtoul(*++argv));

    return 0;
}
